"""
Context.
"""
from inspect import _empty, signature
from typing import Any

from ixoncdkingress.function.api_client import ApiClient
from ixoncdkingress.function.document_db_client import DocumentDBClient


class FunctionResource:
    """
    Describes an IXAPI resource.
    """
    public_id: str
    name: str
    custom_properties: dict[str, Any]
    permissions: set[str] | None

    def __init__(
            self,
            public_id: str,
            name: str,
            custom_properties: dict[str, Any],
            permissions: set[str] | None
        ) -> None:
        self.public_id = public_id
        self.name = name
        self.custom_properties = custom_properties
        self.permissions = permissions

    def __repr__(self) -> str:
        return (
            '<FunctionResource'
            f' public_id={self.public_id},'
            f' name={self.name},'
            f' custom_properties={self.custom_properties!r},'
            f' permissions={self.permissions!r},'
            f'>'
        )

class FunctionContext:
    """
    The context for a Cloud Function.
    """
    config: dict[str, str]
    api_client: ApiClient
    document_db_client: DocumentDBClient | None = None
    user: FunctionResource | None = None
    company: FunctionResource | None = None
    asset: FunctionResource | None = None
    agent: FunctionResource | None = None

    @property
    def agent_or_asset(self) -> FunctionResource:
        """
        Return either an Agent or an Asset resource, depending on what's available. If both are
        available in the context, returns the Asset resource.
        """
        if self.asset:
            return self.asset

        assert self.agent
        return self.agent

    def __init__(
            self,
            config: dict[str, str],
            api_client: ApiClient,
            document_db_client: DocumentDBClient | None = None,
            user: FunctionResource | None = None,
            company: FunctionResource | None = None,
            asset: FunctionResource | None = None,
            agent: FunctionResource | None = None,
            **kwargs: Any
        ) -> None:
        del kwargs

        self.config = config
        self.api_client = api_client
        self.document_db_client = document_db_client
        self.user = user
        self.company = company
        self.asset = asset
        self.agent = agent

    def __repr__(self) -> str:
        return (
            f'<FunctionContext'
            f' config={self.config!r},'
            f' api_client={self.api_client!r},'
            f' document_db_client={self.document_db_client!r},'
            f' user={self.user!r},'
            f' company={self.company!r},'
            f' asset={self.asset!r},'
            f' agent={self.agent!r},'
            f'>'
        )

    @staticmethod
    def expose(function: Any) -> Any:
        """
        Decorator to mark a function as an exposed endpoint.
        """
        sig = signature(function)

        if not sig.parameters:
            raise Exception('Function has no argument for FunctionContext')

        # If the first function argument has a type annotation it should be of FunctionContext
        context_param = sig.parameters[next(iter(sig.parameters))]
        if (context_param.annotation is not _empty
                and context_param.annotation is not FunctionContext):
            raise Exception('First function parameter should be of type FunctionContext')

        function.exposed = True

        return function



import warnings

warnings.warn(
    "ixoncdkingress.cbc had been deprecated, please use ixoncdkingress.function",
    DeprecationWarning,
)

from ixoncdkingress.function.objectstorage.types import ( # noqa: E402, F401, I001
    ListPathResponse,
    PathData,
    PathMapping,
    PathResponse,
    ResourceType,
)



"""
Utility functions.
"""
import json
from time import sleep
from typing import TYPE_CHECKING, Any, Optional
from urllib.parse import parse_qs

from docker import errors, from_env

from ixoncdkingress.function.api_client import ApiClient, Query
from ixoncdkingress.function.context import FunctionContext, FunctionResource
from ixoncdkingress.function.document_db_client import DocumentDBAuthentication, DocumentDBClient
from ixoncdkingress.types import FunctionArguments, FunctionLocation
from ixoncdkingress.webserver.config import Config

if TYPE_CHECKING:
    from docker.models.containers import Container

def read_qs_as_dict(in_put: bytes) -> dict[str, str]:
    """
    Reads and parses the URL query string, returns it as a dict with for each value only the first
    occurrence.
    """
    body = in_put.decode('utf-8')

    # parse_qs always returns non-empty lists!
    return {key: value[0] for (key,value) in parse_qs(body).items()}

def parse_function_location(in_put: str) -> FunctionLocation:
    """
    Parses the full function path into a module path and a function name.
    """

    function_path = in_put.split('.')

    function_name = function_path.pop()
    module_name = '.'.join(function_path)

    if not function_path:
        return 'functions', function_name

    return module_name, function_name

def _parse_context_values(context_items: dict[str, Any]) -> dict[str, Any]:
    """
    Parses expected context_items into FunctionResources
    and discards any unexpected context_items.
    """
    context_values = {}
    for context_name, context_resource in context_items.items():
        # context_values.apiApplication and .accessToken will be sent to the FunctionContext
        # but are unused, they are only used to create the ApiClient.
        if (context_name not in
                ['user', 'company', 'agent', 'asset', 'apiApplication', 'accessToken']):
            continue

        if context_resource is None:
            continue

        # accessToken is not a FunctionResource because it doesn't have
        # a publicId, name and custom field.
        if context_name == 'accessToken':
            context_values[context_name] = context_resource.get('headerValue')
            continue

        permissions = None
        if context_name in ['company', 'agent', 'asset']:
            permissions = set()
            if 'permissions' in context_resource:
                permissions.update(context_resource['permissions'])

        context_values[context_name] = FunctionResource(
            context_resource['publicId'],
            context_resource['name'],
            context_resource['custom'],
            permissions,
        )

    return context_values

def parse_json_input(
        config: Config,
        context_config: dict[str, str],
        body: str
    ) -> tuple[FunctionContext, FunctionLocation, FunctionArguments, str]:
    """
    Parses an application/json request body string into a context, function
    location and function arguments.
    """
    in_put = json.loads(body)

    function_location = parse_function_location(in_put.get('name', ''))

    # Get all context values, may be overwritten if a custom apiApplication is set
    # however this is then still needed for getting authentication values
    context_values = _parse_context_values(in_put.get('context', {}))

    context_config = context_config | in_put.get('config', {})

    api_client_kwargs = {}

    if api_application_res := context_values.get('apiApplication'):
        api_client_kwargs['api_application'] = api_application_res.public_id
    if api_company_res:= context_values.get('company'):
        api_client_kwargs['api_company'] = api_company_res.public_id
    if access_token := context_values.get('accessToken'):
        api_client_kwargs['authorization'] = access_token

    api_client = ApiClient(config.api_client_base_url, **api_client_kwargs)

    # Override the API application if it's set in context_config and not in production mode
    # and refetch custom properties of context_values for scoped custom properties.
    if not config.production_mode and (api_application_res := context_config.get('apiApplication')):
        query: Query = { 'fields': 'custom,name' }
        api_client.set_custom_api_application(api_application_res)

        # Without an access token we don't requery the custom properties
        # as the request will fail anyways.
        if context_values.get('accessToken'):
            api_data = {}
            user_res = api_client.get('MyUser', query=query)
            api_data['user'] = user_res['data']

            # Without a company we don't need to requery these properties
            # as the requests will fail anyways.
            if context_values.get('company'):
                company_res = api_client.get('MyCompany', query=query)
                api_data['company'] = company_res['data']

                if agent := context_values.get('agent'):
                    agent_res = api_client.get(
                        'Agent',
                        { 'publicId': agent.public_id },
                        query=query,
                    )
                    api_data['agent'] = agent_res['data']

                if asset := context_values.get('asset'):
                    asset_res = api_client.get(
                        'Asset',
                        { 'publicId': asset.public_id },
                        query=query,
                    )
                    api_data['asset'] = asset_res['data']

            # Overwrite existing context_values
            context_values = _parse_context_values(api_data)

    document_db_client = None
    if ((db_config := in_put.get('internalConfig', {}).get('dbConfig'))
            and context_values.get('company')
        ):
        # The database name is the public id of the company
        # The collection name is the public id of the Cloud Function template
        document_db_client = DocumentDBClient(
            db_config['uri'],
            context_values['company'].public_id,
            in_put['internalConfig']['publicId'],
            DocumentDBAuthentication(username=db_config['username'], password=db_config['password'])
        )
    elif not config.production_mode and context_values.get('company'):
        document_db_client = DocumentDBClient(
            f'mongodb://localhost:{config.document_db_port}/',
            context_values['company'].public_id,
            'cloud_function_pid',
            DocumentDBAuthentication(username='root', password='mongo-root'), # nosec # noqa: S106
            tls=False
        )

    context = FunctionContext(context_config, api_client, document_db_client, **context_values)

    function_arguments = in_put.get('arguments', {})
    created_on = in_put.get('created_on', '')

    return context, function_location, function_arguments, created_on

def setup_mongo_container(config: Config) -> Optional['Container']:
    """
    Sets up a development MongoDB when not in production mode.

    Will create a docker container named: `ixon-cloud-function-mongo`. The mongodb will be
    accessable at the configured port with user: `root` and password: `mongo-root`

    Returns the created container so that it can be removed on exit.
    """
    if config.production_mode:
        return None

    logger = config.get_logger()
    logger.info('Starting DocumentDB server')

    try:
        client = from_env()

        container: 'Container'
        try:
            container = client.containers.get('ixon-cloud-function-mongo')
            container.remove(force=True)
        except errors.NotFound:
            pass

        container = client.containers.run(
            "mongo:6.0.6-jammy",
            name='ixon-cloud-function-mongo',
            ports={'27017/tcp': config.document_db_port},
            detach=True
        )

        # Ensure database is running and user created. has to be variable because of possible
        # slow development computers of users
        retries, exit_code = 0, 1
        max_retries = 10
        while retries < max_retries and exit_code != 0:
            sleep(1 * retries)
            retries += 1
            exit_code = container.exec_run("""
                mongosh --eval 'db.getSiblingDB("admin").createUser({
                    user: "root",
                    pwd: "mongo-root",
                    roles: [ { role: "userAdminAnyDatabase", db: "admin" } ]
                })'
            """).exit_code

        logger.info('DocumentDB server started successfully')

        return container
    except errors.DockerException as error:
        logger.error("\
            Exception with setting up DocumentDB. \
            Please make sure you have docker installed on your machine \
            and port %s is unused or configure a different port \
            by setting DOCUMENT_DB_PORT to an available port\
        ", config.document_db_port)
        raise error

def shutdown_mongo_container(config: Config, container: 'Container') -> None:
    """
    Tears down and removes the earlier created DocumentDB container
    """
    logger = config.get_logger()

    logger.info('Shutting down DocumentDB server')
    container.remove(force=True)



"""
Module containing a client that is a abstraction of the MongoClient.
"""

from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import Any

from pymongo import MongoClient
from pymongo.collection import Collection
from pymongo.cursor import Cursor
from pymongo.results import (
    DeleteResult,
    InsertManyResult,
    InsertOneResult,
    UpdateResult,
)

TIMEOUT: int = 30000

DocumentType = dict[str, Any]


@dataclass(frozen=True)
class DocumentDBAuthentication:
    """
    Dataclass containing authentication values for the Document DB
    """

    username: str
    password: str


class DocumentDBClient:
    """
    Class for a document db client, this client connects with a document database.
    """

    _mongo_client: MongoClient[DocumentType]
    _database: str
    _default_collection: str
    _current_collection: Collection[DocumentType] | None
    _authentication: DocumentDBAuthentication

    def __init__(
        self,
        connection: str,
        database: str,
        default_collection: str,
        authentication: DocumentDBAuthentication,
        tls: bool = True
    ) -> None:
        """
        database: Company PublicId
        collection: Cloud Function PublicId
        """

        self._database = database
        self._default_collection = default_collection
        self._authentication = authentication

        self._mongo_client = MongoClient(
            connection,
            username=self._authentication.username,
            password=self._authentication.password,
            timeoutMS=TIMEOUT,
            tls=tls
        )

        self._current_collection = None

    def insert_one(self, document: DocumentType) -> InsertOneResult:
        """
        Insert a single document.
        """
        return self._get_collection().insert_one(document)

    def insert_many(self, documents: Iterable[DocumentType]) -> InsertManyResult:
        """
        Insert an iterable of documents.
        """
        return self._get_collection().insert_many(documents)

    def update_one(self, filter_map: Mapping[str, Any], update: Mapping[str, Any]) -> UpdateResult:
        """
        Update a single document matching the filter.
        """
        return self._get_collection().update_one(filter_map, update)

    def update_many(self, filter_map: Mapping[str, Any], update: Mapping[str, Any]) -> UpdateResult:
        """
        Update one or more documents that match the filter.
        """
        return self._get_collection().update_many(filter_map, update)

    def delete_one(self, filter_map: Mapping[str, Any]) -> DeleteResult:
        """
        Delete a single document matching the filter.
        """
        return self._get_collection().delete_one(filter_map)

    def delete_many(self, filter_map: Mapping[str, Any]) -> DeleteResult:
        """
        Delete one or more documents matching the filter.
        """
        return self._get_collection().delete_many(filter_map)

    def find_one(
        self, *args: Any, filter_map: Any | None, **kwargs: Any
    ) -> DocumentType | None:
        """
        Get a single document from the database.
        """
        return self._get_collection().find_one(filter_map, *args, **kwargs)

    def find(self, *args: Any, **kwargs: Any) -> Cursor[DocumentType]:
        """
        Query the database.
        """
        return self._get_collection().find(*args, **kwargs)

    def switch_collection(self, collection_name: str | None) -> None:
        """
        Switch to another collection.

        When collection_name is None or an empty string, the default collection is used
        """
        self._current_collection = None
        self._get_collection(collection_name)

    def _get_collection(self, collection_name: str | None = None) -> Collection[DocumentType]:
        if self._current_collection is not None:
            return self._current_collection

        if collection_name:
            collection_name = f'{self._default_collection}_{collection_name}'

        self._current_collection = self._mongo_client.get_database(self._database).get_collection(
            collection_name or self._default_collection
        )
        return self._current_collection

    def __repr__(self) -> str:
        return (
            '<DocumentDBClient'
            f' database={self._database},'
            f' default_collection={self._default_collection},'
            '>'
        )
